﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Fractarc
{
    public class ArcDef
    {
        public DirPointD Start;
        public DirPointD End;
        public PointD Center;
        public double LengthAngle;
        public double Radius;
        public int Level;
        public enCurveTo CurveTo;

        public static string ToCSVHeader()
        {
            return (DirPointD.ToCSVHeader("Start") + "," + DirPointD.ToCSVHeader("End") + "," + PointD.ToCSVHeader("Center") + "," + "LengthAngle" + "," + "Radius" + "," + "Level" + "," + "CurveTo");
        }
        public string ToCSV()
        {
            return ($"{Start.ToCSV()},{End.ToCSV()},{Center.ToCSV()},{LengthAngle},{Radius},{Level},{CurveTo}");
        }

        public string GCodeMoveTo(FConfig a_config)
        {
            return $"G0 X{Start.X} Y{Start.Y} F{a_config.FeedRateMove}\n";
        }
        public string GCodeMill(FConfig a_config)
        {
            //            return $"G3 X{End.X} Y{End.Y} I{Center.X - Start.X} J{Center.Y - Start.Y}\n";
            string code_type = (CurveTo == enCurveTo.L) ? "G3" : "G2";
            return $"{code_type} X{End.X} Y{End.Y} R{Radius} F{a_config.FeedRateMill}\n";
        }

        public string ToGCode(FConfig a_config)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(GCodeMoveTo(a_config));
            sb.Append(GCodeMill(a_config));
            if (a_config.LineToLineOffset > double.Epsilon)
            {
                double DeltaRadius = a_config.LineToLineOffset;

                for (int i = Level + 1; i < a_config.Levels; i++)
                {
                    ArcDef l1 = CloneInvert();
                    l1.Scale(-DeltaRadius);

                    if (l1.Radius < 0)
                    {
                        Console.WriteLine($"Error - negative radous at level #{Level}. Decrease Levels or increase Radius or decrease LineToLineOffset");
                        break;
                    }
                    sb.Append(l1.GCodeMoveTo(a_config));
                    sb.Append(l1.GCodeMill(a_config));

                    ArcDef l2 = CloneInvert();
                    l2.Scale(DeltaRadius);
                    sb.Append(l2.GCodeMoveTo(a_config));
                    sb.Append(l2.GCodeMill(a_config));

                    DeltaRadius += a_config.LineToLineOffset;
                }
            }

            return sb.ToString();            
        }

        private ArcDef Clone()
        {
            ArcDef arc = new ArcDef();
            arc.Start = Start.Clone();
            arc.End = End.Clone();
            arc.Center = Center.Clone();
            arc.LengthAngle = LengthAngle;
            arc.Radius = Radius;
            arc.Level = Level;
            arc.CurveTo = CurveTo;

            return arc;
        }
        private ArcDef CloneInvert()
        {
            ArcDef arc = new ArcDef();
            arc.Start = End.Clone();
            arc.End = Start.Clone();
            arc.Center = Center.Clone();
            arc.LengthAngle = LengthAngle;
            arc.Radius = Radius;
            arc.Level = Level;
            arc.CurveTo = (CurveTo == enCurveTo.L) ? enCurveTo.R : enCurveTo.L; // invert

            return arc;
        }
        private void Scale(double DeltaRadius)
        {
            double scalefactor = DeltaRadius / Radius;

            double dx;
            double dy;
            // Scale Start
            dx = Start.X - Center.X;
            dy = Start.Y - Center.Y;

            dx *= scalefactor;
            dy *= scalefactor;

            Start.X += dx;
            Start.Y += dy;

            // Scale End
            dx = End.X - Center.X;
            dy = End.Y - Center.Y;

            dx *= scalefactor;
            dy *= scalefactor;

            End.X += dx;
            End.Y += dy;


            // Update Radius
            Radius += DeltaRadius;
        }

    }

}
