﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using YamlDotNet.Serialization;

namespace Fractarc
{
    public class FEngine
    {
        private List<ArcDef> m_arcs;
        FConfig m_config;
        double[] m_scale;
        double[] m_radius;

        public int Levels { get { return m_config.Levels;  } }

        public FEngine(string path_to_settings)
        {
            Deserializer d = new Deserializer();
            using (StreamReader sr = new StreamReader(path_to_settings))
            {
                m_config = d.Deserialize<FConfig>(sr);
            }
            m_scale = new double[m_config.Levels];
            m_radius = new double[m_config.Levels];

            double scale = 1.0;
            double radius = m_config.StartRadius;

            for (int i=0; i<m_config.Levels;i++)
            {
                m_scale[i] = scale;
                m_radius[i] = radius;
                scale *= m_config.Scale;
                radius *= m_config.Scale;
            }


        }

        public List<ArcDef> Render()
        {
            m_arcs = new List<ArcDef>();

            Render1Level(0, m_config.StartPoint);

            return m_arcs;
        }


        private void Render1Level(int a_level, DirPointD dirPoint)
        {
            DirPointD pt = new DirPointD();
            pt.CloneFrom(dirPoint);

            foreach (FSegment fs in m_config.Segments)
            {
                pt = Render1Segment(a_level,pt, fs.LengthAngle,fs.CurveTo);
                DirPointD ch = new DirPointD();
                ch.CloneFrom(pt);
                ch.Dir += fs.BranchAngle;
                if (a_level + 1 < m_config.Levels)
                {
                    Render1Level(a_level + 1, ch);
                }
            }
        }

        private DirPointD Render1Segment(int a_level, DirPointD start, double a_lengthAngle, enCurveTo CurveTo)
        {
            // rotate

            ArcDef arc = new ArcDef();

            arc.CurveTo = CurveTo;
            arc.Start = start;

            arc.Radius = m_radius[a_level];
            arc.LengthAngle = a_lengthAngle;

            arc.Center = new PointD();

            double s_radians = start.Dir * Math.PI / 180.0;

            if (CurveTo == enCurveTo.L)
            {
                arc.Center.X = start.X - arc.Radius * Math.Cos(s_radians);
                arc.Center.Y = start.Y + arc.Radius * Math.Sin(s_radians);
            }
            else
            {
                arc.Center.X = start.X + arc.Radius * Math.Cos(s_radians);
                arc.Center.Y = start.Y - arc.Radius * Math.Sin(s_radians);
            }
            arc.Level = a_level;

            arc.End = RotateDirPoint(arc.Start, arc.Center, a_lengthAngle, CurveTo);


            m_arcs.Add(arc);

            return arc.End;
        }
         static DirPointD RotateDirPoint(DirPointD pointToRotate, PointD centerPoint, double a_angleInDegrees, enCurveTo CurveTo)
        {
            double angleInDegrees = (CurveTo == enCurveTo.L) ? a_angleInDegrees : -a_angleInDegrees;

            double angleInRadians = angleInDegrees * (Math.PI / 180);
            double cosTheta = Math.Cos(angleInRadians);
            double sinTheta = Math.Sin(angleInRadians);
            return new DirPointD
            {
                X =
                    (cosTheta * (pointToRotate.X - centerPoint.X) -
                    sinTheta * (pointToRotate.Y - centerPoint.Y) + centerPoint.X),
                Y =
                    (sinTheta * (pointToRotate.X - centerPoint.X) +
                    cosTheta * (pointToRotate.Y - centerPoint.Y) + centerPoint.Y),
                Dir = pointToRotate.Dir - angleInDegrees
            };
        }

        public string ToGCode()
        {
            StringBuilder sb = new StringBuilder();
            sb.Append($"G17 G90\nG21\n");
            sb.Append($"M4 S10000\n");

            foreach (ArcDef arc in m_arcs)
            {
                sb.Append(arc.ToGCode(m_config));
            }
            sb.Append($"M5 S0\n");
            return sb.ToString();

        }

    }
}
